import math
import re
from typing import Any, List, Tuple, Union

import numpy as np
import pandas as pd

from dbqrqa.utils import DATA, function, remove_nan as clean_nan


def _const_val(source: str, val: str, step: DATA = None) -> float:
    if source == 'custom':
        val = val.strip().replace(' ', '').lower()
        
        if len(val) == 0 or val == '-' or val == 'nan':
            val = np.nan
        
        elif val[-1] == '%':
            val = float(val[:-1])
        
        elif val[-1] in 'kmbt':
            mul = 1e3 ** ('kmbt'.index(val[-1]) + 1)
            val = float(val[:-1]) * mul
        
        else:
            val = float(val)
    
    else:
        if (isinstance(step, pd.Series)):
            val = step.iloc[0]
        
        else:
            val = step.iloc[0, 0]
    
    return val


def _const_loc(df: DATA, idx: str, columns: str) -> DATA:
    idxs = pd.IndexSlice
    if df.index.nlevels > 1: idx = eval(f'idxs[{idx}]')

    if isinstance(df, pd.Series):
        return eval(f'df.loc[{idx}]')
    
    else:
        if df.columns.nlevels > 1: 
            columns = eval(f'idxs[{columns}]')

        return eval(f'df.loc[{idx}, {columns}]')


@function(output_df=False)
def loc(
    df: DATA,
    idx: str = ':',
    columns: str = ':') -> DATA:
    """Select"""
    idxs = pd.IndexSlice
    if df.index.nlevels > 1: idx = eval(f'idxs[{idx}]')

    if isinstance(df, pd.Series):
        do = eval(f'df.loc[{idx}]')

    else:
        if df.columns.nlevels > 1: 
            columns = eval(f'idxs[{columns}]')

        do = eval(f'df.loc[{idx}, {columns}]')

    return clean_nan(do, verbose=False)


@function(output_df=False)
def iloc(
    df: DATA,
    idx: str = ':',
    columns: str = ':') -> DATA:
    """Select"""
    if isinstance(df, pd.Series):
        do = eval(f'df.iloc[{idx}]')

    else:
        do = eval(f'df.iloc[{idx}, {columns}]')

    return clean_nan(do, verbose=False)
    

@function
def loc_filter(
    df: DATA,
    axis: str = 'rows',
    operation: str = 'all',
    condition: str = 'equal',
    source: str = 'custom',
    val: str = None,
    step: DATA = None,
    idx: str = ':',
    columns: str = ':') -> DATA:
    """Filter the data by a condition"""

    val = _const_val(source, val, step)
    axis = 'index' if axis == 'rows' else 'columns'

    idxs = pd.IndexSlice
    if df.index.nlevels > 1: idx = eval(f'idxs[{idx}]')

    if isinstance(df, pd.Series):
        ds = eval(f'df.loc[{idx}]')

    else:
        if df.columns.nlevels > 1: 
            columns = eval(f'idxs[{columns}]')

        ds = eval(f'df.loc[{idx}, {columns}]')

    if condition == 'equal': dt = ds == val
    elif condition == 'gt': dt = ds > val
    elif condition == 'lt': dt = ds < val
    elif condition == 'ge': dt = ds >= val
    elif condition == 'le': dt = ds <= val

    if (isinstance(df, pd.Series)):
        return df[dt], df.index[dt]

    else:
        dts = getattr(dt, operation)(axis=axis)
        do = df.T[dts].T if axis == 'index' else df[dts]
        return do, do.index


@function(output_df=False)
def merge(
    df: DATA,
    dfs: List[DATA],
    axis: str = 'rows',
    ignore_index: bool = False,
    replace_columns: bool = True) -> DATA:
    """Concatenate data frames"""

    if isinstance(df, pd.Series) or \
        any(isinstance(d, pd.Series) for d in dfs):

        if isinstance(df, pd.DataFrame):
            df = df.iloc[:,0]

        for i in range(len(dfs)):
            if isinstance(dfs[i], pd.DataFrame):
                dfs[i] = dfs[i].iloc[:,0]

        return pd.concat([df] + dfs, ignore_index=ignore_index)

    else:
        if axis == 'rows':
            if replace_columns:
                for d in dfs:
                    d.columns = df.columns

            return pd.concat([df] + dfs, axis=0, ignore_index=ignore_index)
        
        else: 
            return pd.concat([df] + dfs, axis=1, ignore_index=ignore_index)


@function
def intersect(
    df: DATA,
    dfs: List[DATA],
    axis: str = 'rows',
    level: int = None) -> DATA:
    """Intersect data frames"""
    idx = df.index

    if level is not None:
        didx = df.index.to_list()

        for d in dfs:
            if isinstance(d, list) or isinstance(d, tuple):
                refs = d

            elif isinstance(d.index[0], str):
                refs = d.index.to_list()
            
            else:
                refs = [e[level] for e in d.index.to_list()]

            fidx = []

            for drx in didx:
                if drx[level] in refs:
                    fidx.append(drx)
            
            didx = fidx

        idx = pd.MultiIndex.from_tuples(didx)
    
    else:
        for d in dfs:
            idx = idx.intersection(d.index)
    
    return df.loc[idx], idx


@function
def union(
    df: DATA,
    dfs: List[DATA]) -> DATA:
    """Union data frames"""
    idx = df.index
    do = pd.concat([df] + dfs)

    for d in dfs:
        idx = idx.union(d.index)
    
    return do.loc[idx], idx


@function
def exclude(
    df: DATA,
    dfs: List[DATA]) -> DATA:
    """Exclude rows from a dataframe"""
    idx = df.index

    for d in dfs:
        idx = idx.drop(d.index)
    
    return df.loc[idx], idx


@function(output_df=False)
def headers(
    df: DATA,
    axis: str = 'rows',
    level: int = None) -> list:
    """Return the headers"""

    def group(idx: list) -> list:
        output = []

        for l in range(len(idx[0])):
            items = set()

            for i in range(len(idx)):
                items.add(idx[i][l])
            
            output.append(list(items))
        
        return output

    if isinstance(df, pd.Series) or axis == 'rows':
        if df.index.nlevels == 1:
            return df.index.tolist()

        else:
            if level is None:
                return group(df.index.tolist())

            else:
                return group(df.index.tolist())[level]
    
    else:
        if df.columns.nlevels == 1:
            return [h[0] for h in df.columns.tolist()]

        else:
            if level is None:
                return group(df.columns.tolist())

            else:
                return group(df.columns.tolist())[level]
            

@function(output_df=False)
def tolist(df: DATA) -> list:
    """Convert the data frame into a list"""
    return df.values.tolist()


@function
def abs(df: DATA) -> DATA:
    """Calculate the absolute values"""
    return df.abs(), df.index


@function
def sum(
    df: DATA, 
    axis: str = 'rows',
    level: int = None,
    keep_col: bool = False) -> DATA:
    """Sum"""
    if isinstance(df, pd.Series): 
        if level is None:
            ds = pd.Series([df.sum(),], index=(('Sum',),))
            return ds, ds.index

        else:
            ds = df.groupby(level=level).sum()
            return ds, ds.index
    
    elif level is None:
        if axis == 'rows': do, idx = df.sum(axis=0), df.columns
        else: do, idx = df.sum(axis=1), df.index
    
    else:
        if axis == 'rows': 
            do, idx = df.groupby(level=level, axis=0).sum(), df.columns
        
        else: 
            do, idx = df.groupby(level=level, axis=1).sum(), df.index

    if keep_col and isinstance(do, pd.Series) or isinstance(do, tuple):
        return pd.DataFrame(do, columns=(('Sum',),)), idx

    else:
        return do, idx


@function
def diff(df: DATA, axis: str = 'rows') -> DATA:
    """Subtract"""
    if isinstance(df, pd.Series): 
        return df.diff().iloc[1:], df.index[1:]
    
    elif axis == 'rows': 
        return df.diff(axis=0).iloc[1:, :], df.columns[1:]
    
    else: 
        return df.diff(axis=1).iloc[:, 1:], df.index[1:]


@function
def div(df: DATA, axis: str = 'rows', keep_col: bool = False) -> DATA:
    """Divide"""
    if keep_col:
        if isinstance(df, pd.Series): 
            dt = df.copy()

            for i in range(1, len(df.index)):
                dt.iloc[i] = df.iloc[i] / df.iloc[i - 1]

            return dt[1:], dt.index[1:]

        elif axis == 'rows': 
            dt = df.copy()

            for i in range(1, len(df.columns)):
                dt.iloc[:, i] = df.iloc[:, i] / df.iloc[:, i - 1]
            
            return dt.iloc[:, 1:], dt.columns[1:]

        else: 
            dt = df.copy()

            for i in range(1, len(df.index)):
                dt.iloc[i, :] = df.iloc[i, :] / df.iloc[i - 1, :]
            
            return dt.iloc[1:, :], dt.index[1:]
        
    else:
        if isinstance(df, pd.Series): 
            return df.iloc[1] / df.iloc[0], (('Divide',),)

        elif axis == 'rows': 
            return df.iloc[:, 1] / df.iloc[:, 0], df.columns[1:]

        else: 
            return df.iloc[1, :] / df.iloc[0, :], df.index[1:]


@function
def mean(
    df: DATA, 
    axis: str = 'rows',
    level: int = None,
    keep_col: bool = False,
    skipna: bool = True) -> DATA:
    """Average"""
    if level is None:
        if isinstance(df, pd.Series): 
            return df.mean(skipna=skipna), (('Average',),)

        else:
            if axis == 'rows': 
                do, idx = df.mean(axis=0, skipna=skipna), df.columns

            else: 
                do, idx = df.mean(axis=1, skipna=skipna), df.index
    
    else:
        if isinstance(df, pd.Series): 
            return df.groupby(level=level).mean(), (('Average',),)
    
        else:
            if axis == 'rows': 
                do = df.groupby(level=level, axis=0).mean()
                idx = df.columns
            
            else: 
                do = df.groupby(level=level, axis=1).mean()
                idx = df.index

    if keep_col and isinstance(do, pd.Series) or isinstance(do, tuple):
        return pd.DataFrame(do, columns=(('Average',),)), idx

    else:
        return do, idx


@function
def sort(
    df: DATA, 
    by: Union[str, List[Union[str, Tuple[str]]]] = None,
    axis: str = 'rows',
    ascending: bool = False) -> DATA:
    """Sort"""
    if isinstance(df, pd.Series) or by is None:
        df = df.sort_values(ascending=ascending)
        return df, df.index

    if len(by) == 0: return df, df.index
    elif isinstance(by[0], list): by = [tuple(x) for x in by]

    by = eval(by) if isinstance(by, str) else list(by)
    if axis == 'rows': df = df.sort_values(by=by, axis=0, ascending=ascending)
    else: df = df.sort_values(by=by, axis=1, ascending=ascending)
    
    return df, df.index


@function
def k_end(
    df: DATA,
    k: Union[int, str],
    direction: str = 'first',
    axis: str = 'rows',
    reverse: bool = False) -> DATA:
    """Select first results"""
    if isinstance(k, str): 
        k = k.strip().replace(' ', '').lower()
        
        if re.match(r'^[0-9]+$', k):
            k = int(k)
        
        elif re.match(r'^[0-9]+[kmbt]$', k):
            mul = 1e3 ** ('kmbt'.index(k[-1]) + 1)
            k = int(int(k[:-1]) * mul)
        
        elif re.match(r'^[0-9]+\.*[0-9]*%$', k):
            if axis == 'rows': size = df.shape[1]
            else: size = df.shape[0]
            k = int(size * float(k[:-1]) / 100)

        else:
            raise RuntimeError('Invalid input')

    if reverse:
        k = -k
        direction = 'last' if direction == 'first' else 'first'

    if direction == 'first':
        if isinstance(df, pd.Series): return df.iloc[:k], df.index
        elif axis == 'rows': return df.iloc[:, :k], df.index
        else: return df.iloc[:k, :], df.index

    elif direction == 'last':
        if isinstance(df, pd.Series): return df.iloc[-k:], df.index
        elif axis == 'rows': return df.iloc[:, -k:], df.index
        else: return df.iloc[-k:, :], df.index


@function
def nth(
    df: DATA,
    n: int,
    direction: str = 'first',
    axis: str = 'rows') -> DATA:
    """Select the nth item"""
    mul = 1 if direction == 'first' else -1
    adj = 0 if direction == 'first' else -1

    if isinstance(df, pd.Series):
        val = df.iloc[(n - 1) * mul + adj]
        df = pd.Series([val,], index=[df.index[(n - 1) * mul + adj],])
        return df, df.index
    
    elif axis == 'rows': 
        dt = df.iloc[:, (n - 1) * mul + adj]
        df = pd.DataFrame(dt)
        return df, df.index

    else: 
        dt = df.iloc[(n - 1) * mul + adj, :]
        df = pd.DataFrame(dt).T
        return df, df.index
    
@function
def compare(
    df: DATA,
    axis: str = 'rows',
    greater: str = 'greater', 
    lower: str = 'lower', 
    equal: str = 'equal') -> str:
    """Compare the values, answer "greater", "lower" or "equal\""""

    def check(value: float) -> str:
        if value > 0: return greater
        elif value < 0: return lower
        elif value == 0: return equal
        else: return np.nan

    if isinstance(df, pd.Series):
        ds = df.diff().iloc[1:]
        return ds.map(check), df.index
    
    else:
        if axis == 'rows': 
            ds = df.diff(axis=0).iloc[1:, :]
            return ds.applymap(check), df.columns

        else: 
            ds = df.diff(axis=1).iloc[:, 1:]
            return ds.applymap(check), df.index
        

@function
def count(
    df: DATA,
    axis: str = 'rows') -> pd.Series:
    """Return the dimension of the data frame"""

    if isinstance(df, list): size = len(df)
    elif isinstance(df, pd.Series): size = len(df)
    elif axis == 'rows': size = len(df.columns)
    elif axis == 'columns': size = len(df.index)

    ds = pd.Series([size], index=(('Count',),))
    return ds, ds.index
        

@function(output_df=False)
def equal(
    df: DATA,
    axis: str = 'rows',
    yes: str = 'yes', 
    no: str = 'no') -> str:
    """Check whether the two values are equal, answer "yes" or "no\""""
    
    def check(value: float) -> str:
        if value == 0: return yes
        elif math.isnan(value): return np.nan
        else: return no

    if isinstance(df, pd.Series):
        ds = df.diff().iloc[1:]
        return ds.map(check), df.index
    
    else:
        if axis == 'rows': 
            ds = df.diff(axis=0).iloc[:, 1:]
            return ds.applymap(check), df.columns

        else: 
            ds = df.diff(axis=1).iloc[1:, :]
            return ds.applymap(check), df.index
        

@function
def stack(
    df: DATA,
    level: int = -1) -> DATA:
    """Stack the data frame"""
    if isinstance(df, pd.Series): return df, df.index

    if isinstance(df.columns.values[0], tuple) or \
        isinstance(df.columns.values[0], list):
        columns = ['_'.join(str(c) for c in col) \
            for col in df.columns.values]
    
    else:
        columns = df.columns.values
    
    df = df.T.reset_index(drop=True).T
    df.columns = columns
    df = df.stack(level=level)
    return df, df.index


@function
def replace(
    df: DATA,
    value: Any,
    target: Any):

    if value == 'nan':
        df = df.fillna(target)

    else:
        df = df.fillna(float('inf'))
        df[df != float('inf')] = target
        df[df == float('inf')] = np.nan
    
    return df, df.index


@function
def cond(
    df: DATA,
    condition: str = 'equal',
    cond_source: str = 'custom',
    cond_val: str = None,
    cond_step: DATA = None,
    target_source: str = 'custom',
    target_val: str = None,
    target_step: DATA = None) -> DATA:
    """Replace the values by a condition"""
    
    cond_val = _const_val(cond_source, cond_val, cond_step)
    target_val = _const_val(target_source, target_val, target_step)

    if condition == 'equal': df[df == cond_val] = target_val
    elif condition == 'gt': df[df > cond_val] = target_val
    elif condition == 'lt': df[df < cond_val] = target_val
    elif condition == 'ge': df[df >= cond_val] = target_val
    elif condition == 'le': df[df <= cond_val] = target_val
    elif condition == 'ne': df[df != cond_val] = target_val

    return df, df.index


@function
def transpose(
    df: DATA) -> DATA:
    """Transpose the data frame"""

    if isinstance(df, pd.Series):
        ds = pd.DataFrame([df.values,], columns=df.index)

    else:
        ds = df.T

    return ds, ds.index


@function
def const_add(
    df: DATA,
    idx: str = ':',
    columns: str = ':',
    source: str = 'custom',
    val: str = None,
    step: DATA = None,
    table: DATA = None) -> DATA:
    """Add a constant to the data frame"""
    if source == 'table':
        df = df.add(table)
    
    else:
        val = _const_val(source, val, step)
        dt = _const_loc(df, idx, columns) + val
        df.update(dt)

    return df, df.index


@function
def const_subtract(
    df: DATA,
    idx: str = ':',
    columns: str = ':',
    source: str = 'custom',
    val: str = '0',
    step: DATA = None) -> DATA:
    """Stract a constant from the data frame"""
    val = _const_val(source, val, step)
    dt = _const_loc(df, idx, columns) - val
    df.update(dt)
    return df, df.index


@function
def const_mul(
    df: DATA,
    idx: str = ':',
    columns: str = ':',
    source: str = 'custom',
    val: str = '1',
    step: DATA = None) -> DATA:
    """Multiply a constant to the data frame"""
    val = _const_val(source, val, step)
    dt = _const_loc(df, idx, columns) * val
    df.update(dt)
    return df, df.index


@function
def const_div(
    df: DATA,
    idx: str = ':',
    columns: str = ':',
    source: str = 'custom',
    val: str = '1',
    step: DATA = None) -> DATA:
    """Divide the data frame by a constant"""
    val = _const_val(source, val, step)
    dt = _const_loc(df, idx, columns) / val
    df.update(dt)
    return df, df.index


@function
def remove_nan(df: DATA, axis: str = 'rows') -> DATA:
    """Remove rows with all null values"""

    if axis == 'rows':
        cols = []

        for c in range(len(df.columns)):
            remove = True

            for r in range(len(df)):
                if not pd.isnull(df.iloc[r, c]):
                    remove = False
            
            if remove:
                cols.append(c)

        dt = df.copy().drop(cols, axis=1)
        
    else:
        rows = []

        for r in range(len(df)):
            remove = True

            for c in range(len(df.columns)):
                if not pd.isnull(df.iloc[r, c]):
                    remove = False
            
            if remove:
                rows.append(r)
        
        dt = df.copy().drop(rows, axis=0)
    
    return dt, dt.index


@function(input_df=False, output_df=False)
def choices(choice: str = 'yes') -> str:
    """Choose of choices"""
    return choice


@function(output_df=False)
def quantify(df: DATA, unit: str) -> str:
    """Quantify in unit"""
    if isinstance(df, pd.Series): num = df.iloc[0]
    else: num = df.iloc[0, 0]
    
    if unit == 'percent': num *= 100
    elif unit == 'thousands': num /= 1e3
    elif unit == 'millions': num /= 1e6
    elif unit == 'billions': num /= 1e9
    elif unit == 'trillions': num /= 1e12
    else: raise KeyError(f'Unknown unit: {unit}')

    return '%.2f %s' % (num, unit)


@function(output_df=False)
def reverse(df: DATA, axis: str = 'rows') -> DATA:
    if isinstance(df, pd.Series): return df.iloc[::-1]
    elif axis == 'rows': return df.iloc[:,::-1]
    else: return df.iloc[::-1]


@function(input_df=False, output_df=False)
def empty(text: str = 'None') -> str:
    """Indicate that the answer is empty"""
    return text


@function(input_df=False, output_df=False)
def ambiguous(text: str = 'terminate:ambiguous') -> str:
    """Indicate that the answer is ambiguous"""
    return text


@function(input_df=False, output_df=False)
def invalid_data(text: str = 'terminate:invalid_data') -> str:
    """Indicate that the queries are invalid"""
    return text


@function(input_df=False, output_df=False)
def other_problems(text: str = 'terminate:unsolvable') -> str:
    """Indicate that the answer is unsolvable"""
    return text



from typing import Any, Callable, List, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

DATA = Union[str, list, pd.DataFrame, pd.Series]


def todf(
    object: Union[float, pd.DataFrame, pd.Series],
    index: List[str]) -> DATA:

    if isinstance(object, float):
        return pd.Series([object], index=index[0])
    
    elif isinstance(object, pd.DataFrame) or isinstance(object, pd.Series):
        if len(object.index) == 0:
            raise RuntimeError('Empty data frame')
    
    return object


def function(
    _func: Callable = None,
    *,
    input_df: bool = True,
    output_df: bool = True):

    def _function(func):
        def wrapper(data: Any, *args, spec: bool = False, **kwargs):
            if spec:
                return func
            
            else:
                if isinstance(data, pd.DataFrame) \
                    or isinstance(data, pd.Series):
                    data = data.copy()

                if output_df:
                    output, idx = func(data, *args, **kwargs)
                    return todf(output, idx)
            
                else:
                    if input_df:
                        return func(data, *args, **kwargs)

                    else:
                        return func(*args, **kwargs)
        
        return wrapper
    
    if _func is None:
        return _function

    else:
        return _function(_func)


def remove_nan(df: pd.DataFrame, verbose: bool = True) -> pd.DataFrame:
    if not isinstance(df, pd.DataFrame):
        return df

    include, idx = [], []
    nulls = df.isnull().sum(axis=1)

    for r in tqdm(range(
        len(df.index)), desc='Cleaning', disable=not verbose):

        if nulls[r] < len(df.columns):
            include.append(r)
            idx.append(df.index[r])
    
    dt = pd.DataFrame(
        data=np.empty((len(include), len(df.columns))) * np.nan,
        index=pd.MultiIndex.from_tuples(idx),
        columns=df.columns)

    for r in include:
        dt.loc[df.index[r]] = df.iloc[r]

    return dt



import argparse
import json
from os.path import join

from dbqrqa.dataset import TableDataset
from dbqrqa.evaluation import evaluate

DEFAULT_OPENAI_KEY_PATH = join('data', 'keys', 'openai.txt')
DEFAULT_EVALUATOR = 'gpt-binary'
DEFAULT_MODEL = 'gpt-4o'
DEFAULT_RETRY = 3
DEFAULT_BACKUP_PATH = join('data', 'backup')
DEFAULT_DATA_PATH = 'data'


def main(args: argparse.Namespace) -> None:
    """
    The GPT evaluator is not part of the standard unit test
    due to the cost and security considerations.

    Call this special test function with an API key to test
    the evaluator.
    """

    dataset = TableDataset()
    practice = dataset.practice

    with open(args.key) as file:
        openai_key = file.read().strip()

    with open(join(args.data, 'tests', 'evaluation', 'answers.json')) as file:
        answers = json.load(file)

    accuracy, _ = evaluate(practice.questions, answers, practice.labels,
        args.evaluator, args.model, args.retry, openai_key, args.backup)
    
    print('Custom implementation test accuracy:', accuracy)

    practice.answers = answers
    accuracy, _ = practice.evaluate(args.evaluator, args.model, 
        args.retry, openai_key, args.backup)

    print('TableDataset implementation test accuracy:', accuracy)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--evaluator', type=str, default=DEFAULT_EVALUATOR,
        help='Evaluation method', choices=('gpt-binary', 'gpt-score'))
    parser.add_argument('--model', type=str, default=DEFAULT_MODEL,
        help='OpenAI model')
    parser.add_argument('--retry', type=int, default=DEFAULT_RETRY,
        help='Maximum number of retries for invalid outputs')
    parser.add_argument('--key', type=str, default=DEFAULT_OPENAI_KEY_PATH,
        help='Path to OpenAI API key')
    parser.add_argument('--backup', type=str, default=DEFAULT_BACKUP_PATH,
        help='Backup path for the model\'s outputs and scores')
    parser.add_argument('--data', type=str, default=DEFAULT_DATA_PATH)
    args = parser.parse_args()
    main(args)



import json
import unittest
from os.path import join

from dbqrqa.dataset import TableDataset

TEST_PATH = join('data', 'tests', 'evaluation')


class TableDatasetTestCase(unittest.TestCase):
    def test_dataset(self):
        dataset = TableDataset()

        self.assertGreater(len(dataset.practice.questions), 0)
        self.assertGreater(len(dataset.practice.labels), 0)
        self.assertGreater(len(dataset.practice.queries), 0)
        self.assertGreater(len(dataset.practice.tables), 0)

        with open(join(TEST_PATH, 'answers.json')) as file:
            dataset.practice.answers = json.load(file)

        dataset.practice.answer('chat-1-01', 1, 'lower')
        
        accuracy, _ = dataset.practice.evaluate()
        self.assertEqual(accuracy, 0.82)


if __name__ == "__main__":
    unittest.main()



import json
import unittest
from os import listdir
from os.path import join

from dbqrqa.evaluation import evaluate

QUESTION_PATH = join('data', 'practice', 'questions')
TEST_PATH = join('data', 'tests', 'evaluation')


class EvaluationTestCase(unittest.TestCase):
    def build_test_input(self):
        self.questions = {}
        self.answers = {}
        self.labels = {}

        for chat_id in listdir(QUESTION_PATH):
            questions, labels, answers = {}, {}, {}

            for question_id in range(1, 11):
                chat_file = join(
                    QUESTION_PATH, 
                    chat_id, 
                    'question-%02d.json' % question_id)
            
                with open(chat_file) as file:
                    data = json.load(file)
                
                questions[str(question_id)] = data['question']
                answers[str(question_id)] = data['answer']
                labels[str(question_id)] = data['answer']
            
            self.questions[chat_id] = questions
            self.answers[chat_id] = answers
            self.labels[chat_id] = labels

    def test_heuristic_basic(self):
        self.build_test_input()

        accuracy, _ = evaluate(
            self.questions, self.answers, self.labels)
        
        self.assertEqual(accuracy, 1.0)

    def test_heuristic_noise(self):
        self.build_test_input()

        with open(join(TEST_PATH, 'answers.json')) as file:
            self.answers = json.load(file)
        
        accuracy, _ = evaluate(
            self.questions, self.answers, self.labels)
        
        self.assertEqual(accuracy, .82)


if __name__ == "__main__":
    unittest.main()



import json
import pickle
from os import listdir
from os.path import isdir, join
from typing import Any, Dict, Union

import pandas as pd

from dbqrqa.evaluation import (
    DEFAULT_EVALUATOR,
    DEFAULT_GPT_MODEL,
    DEFAULT_RETRY,
    OUTPUT_TYPES, 
    evaluate)
from dbqrqa.types import ANSWER_TYPES

STAGES = ('practice', 'train', 'test')

DEFAULT_DATA_PATH = 'data'


class TableSplit:
    """
    A split of the dataset for the practicing, training, or testing stage.
    """

    def __init__(
        self, 
        stage: str, 
        data_path: str = DEFAULT_DATA_PATH):

        self.stage = stage
        self.data_path = data_path
        self.answers: Union[ANSWER_TYPES, None] = None

        self.load()

    def load(self):
        stage_path = join(self.data_path, self.stage)
        question_path = join(stage_path, 'questions')
        table_path = join(stage_path, 'tables')

        self.chats = {}

        for chat_id in listdir(question_path):
            chat = {}

            for question_id in range(1, 11):
                file_path = join(
                    question_path, 
                    chat_id, 
                    'question-%02d.json' % question_id)
                
                with open(file_path) as file:
                    sample = json.load(file)

                file_path = join(
                    table_path,
                    chat_id,
                    'question-%02d.pkl' % question_id)
                
                with open(file_path, 'rb') as file:
                    sample['tables'] = pickle.load(file)
                
                chat[str(question_id)] = sample
        
            self.chats[chat_id] = chat

    def answer(
        self, 
        chat_id: str, 
        question_id: Union[int, str], 
        answer: ANSWER_TYPES):

        # Check chat ID
        chat = chat_id.split('-')

        assert len(chat) == 3, 'Invalid chat ID'
        assert chat[0] == 'chat', 'Invalid chat ID'
        assert chat[1].isnumeric() and len(chat[1]) == 1, 'Invalid chat ID'
        assert chat[2].isnumeric() and len(chat[2]) == 2, 'Invalid chat ID'

        # Check question ID
        question_id = str(question_id)

        assert question_id.isnumeric(), 'Question ID must be an integer'

        # Remove preceeding zeros
        question_id = str(int(question_id))

        if int(question_id) < 1 or int(question_id) > 10:
            raise ValueError('Question ID must be between 1 and 10.')

        if self.answers is None:
            self.answers = {}

        if chat_id not in self.answers:
            self.answers[chat_id] = {}
        
        self.answers[chat_id][question_id] = answer

    def evaluate(
        self, 
        evaluator: str = DEFAULT_EVALUATOR,
        model: str = DEFAULT_GPT_MODEL,
        retry: int = DEFAULT_RETRY,
        openai_key: Union[str, None] = None,
        backup_path: Union[str, None] = None,
        is_notebook: bool = False) -> OUTPUT_TYPES:

        if self.answers is None:
            raise RuntimeError('No answers assigned to this dataset.' \
                'See README file to learn how to assign the answers.')

        return evaluate(self.questions, self.answers, self.labels,
            evaluator=evaluator, model=model, retry=retry,
            openai_key=openai_key, backup_path=backup_path,
            is_notebook=is_notebook)

    def _get_property(self, key: str) -> Dict[str, Dict[str, Any]]:
        chats = {}

        for chat_id, chat in self.chats.items():
            chats[chat_id] = {}

            for question_id, sample in chat.items():
                chats[chat_id][question_id] = sample[key]
        
        return chats

    @property
    def questions(self) -> Dict[str, Dict[str, str]]:
        return self._get_property('question')
    
    @property
    def labels(self) -> ANSWER_TYPES:
        return self._get_property('answer')
    
    @property
    def queries(self) -> Dict[str, Dict[str, str]]:
        return self._get_property('queries')
    
    @property
    def tables(self) -> Dict[str, Dict[str, pd.DataFrame]]:
        return self._get_property('tables')


class TableDataset:
    """
    A dataset class for the project's shared task where table inputs are given
    (no database query required). For more information, visit https://dbqr-qa.github.io
    """

    def __init__(self, data_path: str = DEFAULT_DATA_PATH):
        self.data_path = data_path
        self.practice: Union[TableSplit, None] = None
        self.train: Union[TableSplit, None] = None
        self.test: Union[TableSplit, None] = None
        self.load()

    def load(self):
        for stage in STAGES:
            if isdir(join(self.data_path, stage)):
                setattr(self, stage, TableSplit(stage, self.data_path))



import json
import logging
from typing import Dict, Iterable, List, Tuple, Union
from os.path import isfile, join
from pathlib import Path

from openai import OpenAI
from tqdm import tqdm

from dbqrqa.prompts import (
    GPT_PROMPT_SYSTEM,
    GPT_PROMPT_BINARY, 
    GPT_PROMPT_SCORE)
from dbqrqa.types import (
    VALUE_TYPES,
    ANSWER_TYPES)

OUTPUT_TYPES = Tuple[float, Dict[str, Dict[str, float]]]

EVALUATORS = ('heuristic', 'gpt-binary', 'gpt-score')

DEFAULT_EVALUATOR = 'heuristic'
DEFAULT_GPT_MODEL = 'gpt-4o'
DEFAULT_RETRY = 3


def _match_heuristic(answer: VALUE_TYPES, label: VALUE_TYPES) -> bool:
    """
    Determine the label type (string, int, or float) and convert
    both the answer and label to the same format, then compare.
    """

    if isinstance(label, str):
        return str(answer).lower().strip() == label.lower().strip()

    elif isinstance(label, int):
        try:
            return int(answer) == label

        except (ValueError, TypeError):
            return False
    
    elif isinstance(label, float):
        try:
            return '%.2f' % float(answer) == '%.2f' % label

        except (ValueError, TypeError):
            return False


def evaluate_heuristic(answer: ANSWER_TYPES, label: ANSWER_TYPES) -> int:
    """
    Evaluate a single answer-label pair using the heuristic evaluator.
    For more information, visit https://dbqr-qa.github.io/quickstart.html#heuristic
    """

    if isinstance(answer, set):
        answer = list
    
    if isinstance(label, list):
        if not isinstance(answer, list) or len(answer) != len(label):
            return 0
        
        answer = sorted(answer)
        label = sorted(label)

        for item, ref in zip(answer, label):
            if not _match_heuristic(item, ref):
                return 0

        return 1

    elif isinstance(label, dict):
        if not isinstance(answer, dict) or len(answer) != len(label):
            return 0
        
        for key, ref in label.items():
            if key not in answer:
                return 0

            if not _match_heuristic(answer[key], ref):
                return 0
        
        return 1

    else:
        return int(_match_heuristic(answer, label))


def init_gpt_messages() -> List[Dict[str, str]]:
    """
    Initialize GPT's input by explaning the task.
    """
    return [{'role': 'system', 'content': GPT_PROMPT_SYSTEM}]


def evaulate_gpt(
    question: str,
    answer: ANSWER_TYPES,
    label: ANSWER_TYPES,
    messages: List[Dict[str, str]],
    prompt: str,
    client: OpenAI,
    model: str,
    scoring: str = 'binary',
    retry: int = DEFAULT_RETRY) -> Tuple[float, Dict[str, str]]:
    
    """
    Evaluate a single answer-label pair using the GPT evaluator.
    For more information, visit https://dbqr-qa.github.io/quickstart.html#gpt

    Use init_gpt_messages() to initialize `messages` (GPT's input).
    This function will automatically append the GPT's response to `messages` before
    returning the score and the updated `messages`.

    For binary evaluator, the score can either be zero or one.
    For scoring evaluator, the score ranges from zero to one on 0.1 intervals.

    The `retry` parameter sets the limit of how many times the function
    will try to get the score if the previous attempts failed due to
    invalid GPT's responses.
    """

    def append(messages: List[Dict[str, str]], content: str):
        messages.append({
            'role': 'assistant',
            'content': content})
        
        return messages
    
    if scoring not in ('binary', 'score'):
        raise KeyError('Invalid scoring choice. ' \
            'The valid choices are binary or score.')

    content = prompt.replace('{{question}}', question) \
                .replace('{{answer}}', str(answer)) \
                .replace('{{label}}', str(label))

    messages.append({
        'role': 'user',
        'content': content})

    attempt = 0

    while attempt < retry:
        response = client.chat.completions.create(
            model=model,
            messages=messages)
        
        content = response.choices[0].message.content

        # Extract the JSON output by detecting lines in a code block
        is_json_lines = False
        json_output = ''

        for line in content.split('\n'):
            if '```' in line:
                is_json_lines = not is_json_lines
            
            elif is_json_lines:
                json_output += line + '\n'
        
        try:
            output = json.loads(json_output)

            if scoring == 'binary':
                if output['result'] == 'yes':
                    return 1, append(messages, content)
                
                elif output['result'] == 'no':
                    return 0, append(messages, content)

                else:
                    logging.warning('Invalid GPT output. Retrying.')
                    attempt += 1

            else:
                score = int(output['result'])

                if 0 <= score <= 10:
                    return float(score) / 10, append(messages, content)

                else:
                    logging.warning('Invalid GPT output. Retrying.')
                    attempt += 1
        
        except (json.decoder.JSONDecodeError, KeyError, ValueError):
            logging.warning('Invalid GPT output. Retrying.')
            attempt += 1

    raise RuntimeError('Invalid GPT output exceeded retry limit.')


def evaluate(
    questions: Dict[str, Dict[str, str]],
    answers: ANSWER_TYPES,
    labels: ANSWER_TYPES,
    evaluator: str = DEFAULT_EVALUATOR,
    model: str = DEFAULT_GPT_MODEL,
    retry: int = DEFAULT_RETRY,
    openai_key: Union[str, None] = None,
    backup_path: Union[str, None] = None,
    is_notebook: bool = False) -> OUTPUT_TYPES:

    """
    Heuristic/GPT evaluator.
    For more information, visit https://dbqr-qa.github.io/quickstart.html#evaluation

    Parameters `model`, `retry`, `openai_key`, `backup_path` are not used for
    the heuristic evaluator.

    The `is_notebook` parameter prevents tqdm from outputting 
    a new progress bar for every question.
    """

    def display(
        items: Iterable, 
        desc: Union[str, None] = None, 
        leave: bool = True) -> Iterable:

        if evaluator == 'heuristic' or (is_notebook and not leave):
            return items
    
        else:
            return tqdm(items, desc=desc, leave=leave)

    if evaluator not in EVALUATORS:
        raise KeyError(f'Invalid evaluators. The valid choices are {EVALUATORS}.')
    
    if evaluator.startswith('gpt'):
        if openai_key is None:
            raise RuntimeError("OpenAI API key is required for a GPT evaluator.")
        
        else:
            client = OpenAI(api_key=openai_key)

        scoring = evaluator.split('-')[1]
        prompt = GPT_PROMPT_BINARY if scoring == 'binary' else GPT_PROMPT_SCORE

        if backup_path is None:
            message = 'No backup path provided for the GPT evaluator.' \
                'The output will not be saved, and the evaluation ' \
                'cannot resume if terminated early.'
            
            logging.warning(message)

    scores, messages = {}, {}

    if backup_path is not None:
        Path(backup_path).mkdir(parents=True, exist_ok=True)
        backup_file = join(backup_path, f'gpt-evaluation-scores.json')

        if isfile(backup_file):
            try:
                with open(backup_file) as file:
                    backup = json.load(file)
                    scores = backup['scores']
                    messages = backup['messages']
            
            except json.decoder.JSONDecodeError:
                logging.warning('Unable to read the backup file.')

    for chat_id in display(questions.keys(), desc='Chat'):
        if chat_id not in scores or chat_id not in messages:
            scores[chat_id] = {}

            # Initialize GPT's input by explaining the task.
            messages[chat_id] = init_gpt_messages()

        # All conversations have ten questions.
        for question_id in display(range(1, 11), desc='Question', leave=False):
            question_key = str(question_id)

            if question_key in scores[chat_id]:
                continue

            question = questions[chat_id][question_key]
            answer = answers[chat_id][question_key]
            label = labels[chat_id][question_key]

            if evaluator == 'heuristic':
                scores[chat_id][question_key] = evaluate_heuristic(
                    answer, label)
            
            else:
                scores[chat_id][question_key], messages[chat_id] = evaulate_gpt(
                    question, answer, label, messages[chat_id], 
                    prompt, client, model, scoring, retry)

                if backup_path is not None:
                    with open(backup_file, 'w') as file:
                        json.dump({
                            'scores': scores,
                            'messages': messages
                        }, file, indent=2)

    all_scores = []

    for chat_scores in scores.values():
        all_scores += list(chat_scores.values())
    
    return float(sum(all_scores)) / len(all_scores), scores



import argparse
import json
from os.path import join

from dbqrqa.dataset import TableDataset

DEFAULT_DATA_PATH = 'data'


def main(args: argparse.Namespace) -> None:
    """
    Add noise to the practice labels for testing.
    """

    dataset = TableDataset()
    practice = dataset.practice

    with open(join(args.data, 'tests', 'evaluation', 'noise.json')) as file:
        noise = json.load(file)

    answers = {}

    for chat_id, chat in practice.labels.items():
        answers[chat_id] = {}

        for question_id, label in chat.items():
            answers[chat_id][question_id] = label

            if chat_id in noise and question_id in noise[chat_id]:
                answers[chat_id][question_id] = noise[chat_id][question_id]

    with open(join(args.data, 'tests', 'evaluation', 'answers.json'), 'w') as file:
        json.dump(answers, file, indent=2)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--data', type=str, default=DEFAULT_DATA_PATH)
    args = parser.parse_args()
    main(args)



GPT_PROMPT_SYSTEM = """You are an evaluator. 
Given a series of conversational questions, 
your task is to compare an answer to the last question predicted by an AI 
to an answer labeled by a human."""

GPT_PROMPT_BINARY = """Question:
{{question}}

AI's answer: 
{{answer}}

Human's answer: 
{{label}}

Are the two answers to the last question the same?
Answer "yes" or "no" in the following JSON format:
```
{
  "result": "yes" or "no"
}
```
Do not explain or output anything else."""

GPT_PROMPT_SCORE = """Question:
{{question}}

Compare the following answers to the last question in the above conversation.
AI's answer: 
{{answer}}

Human's answer: 
{{label}}

On a scale of 0 to 10, 0 = not at all and 10 = same, how similar are the two answers?
Answer in the following JSON format:
```
{
  "result": A score from 0 to 10
}
```
Do not explain or output anything else."""
